@@ -1,30 +1,32 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from decimal import Decimal
-
 from trytond import backend
+from trytond.i18n import gettext
 from trytond.model import ModelSQL, ValueMixin, fields
+from trytond.model.exceptions import RequiredValidationError
 from trytond.pyson import Eval
 from trytond.pool import Pool, PoolMeta
-from trytond.transaction import Transaction
 from trytond.tools.multivalue import migrate_property
 
+from trytond.modules.product import round_price
+from trytond.modules.purchase.purchase import (
+    get_shipments_returns, search_shipments_returns)
 
-__all__ = ['PurchaseRequest', 'PurchaseConfig',
-    'PurchaseConfigPurchaseDropLocation', 'Purchase', 'PurchaseLine',
-    'ProductSupplier', 'CreatePurchase', 'PurchaseHandleShipmentException']
 purchase_drop_location = fields.Many2One(
     'stock.location', "Purchase Drop Location", domain=[('type', '=', 'drop')])
 
 
-class PurchaseRequest:
-    __metaclass__ = PoolMeta
+class Request(metaclass=PoolMeta):
     __name__ = 'purchase.request'
 
     customer = fields.Many2One('party.party', 'Customer', readonly=True,
         states={
             'invisible': ~Eval('customer'),
-            })
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     delivery_address = fields.Many2One('party.address', 'Delivery Address',
         domain=[('party', '=', Eval('customer'))],
         states={
@@ -34,8 +36,7 @@ class PurchaseRequest:
         depends=['customer', 'state'])
 
 
-class PurchaseConfig:
-    __metaclass__ = PoolMeta
+class Configuration(metaclass=PoolMeta):
     __name__ = 'purchase.configuration'
 
     purchase_drop_location = fields.MultiValue(purchase_drop_location)
@@ -46,18 +47,16 @@ class PurchaseConfig:
             'purchase_drop_location').default_purchase_drop_location()
 
 
-class PurchaseConfigPurchaseDropLocation(ModelSQL, ValueMixin):
+class ConfigurationPurchaseDropLocation(ModelSQL, ValueMixin):
     "Purchase Configuration Purchase Drop Location"
     __name__ = 'purchase.configuration.purchase_drop_location'
     purchase_drop_location = purchase_drop_location
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
 
-        super(PurchaseConfigPurchaseDropLocation, cls).__register__(
-            module_name)
+        super().__register__(module_name)
 
         if not exist:
             cls._migrate_property([], [], [])
@@ -81,14 +80,17 @@ class PurchaseConfigPurchaseDropLocation(ModelSQL, ValueMixin):
             return None
 
 
-class Purchase:
-    __metaclass__ = PoolMeta
+class Purchase(metaclass=PoolMeta):
     __name__ = 'purchase.purchase'
 
     customer = fields.Many2One('party.party', 'Customer', readonly=True,
         states={
             'invisible': ~Eval('customer'),
-            })
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     delivery_address = fields.Many2One('party.address', 'Delivery Address',
         domain=[('party', '=', Eval('customer'))],
         states={
@@ -96,13 +98,13 @@ class Purchase:
             'invisible': ~Eval('customer'),
             },
         depends=['state', 'customer'])
-    drop_shipments = fields.Function(fields.One2Many('stock.shipment.drop',
-            None, 'Drop Shipments',
+    drop_shipments = fields.Function(fields.Many2Many('stock.shipment.drop',
+            None, None, "Drop Shipments",
             states={
                 'invisible': ~Eval('customer'),
                 },
             depends=['customer']),
-        'get_drop_shipments')
+        'get_drop_shipments', searcher='search_drop_shipments')
     drop_location = fields.Many2One('stock.location', 'Drop Location',
         domain=[('type', '=', 'drop')],
         states={
@@ -111,14 +113,6 @@ class Purchase:
             },
         depends=['customer'])
 
-    @classmethod
-    def __setup__(cls):
-        super(Purchase, cls).__setup__()
-        cls._error_messages.update({
-                'delivery_address_required': ('A delivery address must be '
-                    'defined for quotation of purchase "%s".'),
-                })
-
     @staticmethod
     def default_drop_location():
         pool = Pool()
@@ -141,15 +135,17 @@ class Purchase:
             address = self.delivery_address.full_address
         return address
 
-    def get_drop_shipments(self, name):
-        DropShipment = Pool().get('stock.shipment.drop')
-        return list(set(m.shipment.id for l in self.lines for m in l.moves
-                if isinstance(m.shipment, DropShipment)))
+    get_drop_shipments = get_shipments_returns('stock.shipment.drop')
+    search_drop_shipments = search_shipments_returns('stock.shipment.drop')
 
     def check_for_quotation(self):
         super(Purchase, self).check_for_quotation()
         if self.customer and not self.delivery_address:
-            self.raise_user_error('delivery_address_required', self.rec_name)
+            raise RequiredValidationError(
+                gettext('sale_supply_drop_shipment'
+                    '.msg_delivery_address_required_quotation_purchase') % {
+                    'purchase': self.rec_name,
+                    })
 
     def create_move(self, move_type):
         pool = Pool()
@@ -177,12 +173,11 @@ class Purchase:
             )
 
 
-class PurchaseLine:
-    __metaclass__ = PoolMeta
+class Line(metaclass=PoolMeta):
     __name__ = 'purchase.line'
 
     def get_to_location(self, name):
-        result = super(PurchaseLine, self).get_to_location(name)
+        result = super().get_to_location(name)
         if self.purchase.customer:
             return self.purchase.drop_location.id
         return result
@@ -190,18 +185,15 @@ class PurchaseLine:
     def get_move(self, move_type):
         pool = Pool()
         Currency = pool.get('currency.currency')
-        move = super(PurchaseLine, self).get_move(move_type)
+        move = super().get_move(move_type)
         if move and self.purchase.customer:
-            digits = move.__class__.cost_price.digits
             cost_price = Currency.compute(
                 move.currency, move.unit_price, move.company.currency)
-            move.cost_price = cost_price.quantize(
-                Decimal(str(10.0 ** -digits[1])))
+            move.cost_price = round_price(cost_price)
         return move
 
 
-class ProductSupplier:
-    __metaclass__ = PoolMeta
+class ProductSupplier(metaclass=PoolMeta):
     __name__ = 'purchase.product_supplier'
 
     drop_shipment = fields.Boolean('Drop Shipment',
@@ -213,21 +205,21 @@ class ProductSupplier:
         fields.Boolean("Drop Shipment Available"),
         'on_change_with_drop_shipment_available')
 
-    @fields.depends('product',
-        '_parent_product.type', '_parent_product.supply_on_sale')
+    @fields.depends('product', 'template',
+        '_parent_product.type', '_parent_product.supply_on_sale',
+        '_parent_template.type', '_parent_template.supply_on_sale')
     def on_change_with_drop_shipment_available(self, name=None):
-        if self.product and self.product.type in {'goods', 'assets'}:
-            return self.product.supply_on_sale
+        product = self.product or self.template
+        if product and product.type in {'goods', 'assets'}:
+            return product.supply_on_sale
 
 
-class CreatePurchase:
-    __metaclass__ = PoolMeta
+class RequestCreatePurchase(metaclass=PoolMeta):
     __name__ = 'purchase.request.create_purchase'
 
     @classmethod
     def _group_purchase_key(cls, requests, request):
-        result = super(CreatePurchase, cls)._group_purchase_key(requests,
-            request)
+        result = super()._group_purchase_key(requests, request)
         result += (
             ('customer', request.customer.id if request.customer else None),
             ('delivery_address', request.delivery_address.id
@@ -236,25 +228,22 @@ class CreatePurchase:
         return result
 
 
-class PurchaseHandleShipmentException:
-    __metaclass__ = PoolMeta
+class HandleShipmentException(metaclass=PoolMeta):
     __name__ = 'purchase.handle.shipment.exception'
 
     def transition_handle(self):
         pool = Pool()
         Sale = pool.get('sale.sale')
-        Purchase = pool.get('purchase.purchase')
         Move = pool.get('stock.move')
 
-        super(PurchaseHandleShipmentException, self).transition_handle()
+        super().transition_handle()
 
         sales = set()
         moves = set()
         to_recreate = set(self.ask.recreate_moves)
         domain_moves = set(self.ask.domain_moves)
-        purchase = Purchase(Transaction().context['active_id'])
 
-        for line in purchase.lines:
+        for line in self.record.lines:
             if not set(line.moves) & domain_moves:
                 continue
             if not any(m in to_recreate for m in line.moves):
@@ -266,7 +255,7 @@ class PurchaseHandleShipmentException:
                         sales.add(sale_line.sale)
 
         if moves:
-            Move.cancel(list(moves))
+            Move.cancel(moves)
         if sales:
-            Sale.process(list(sales))
+            Sale.__queue__.process(sales)
         return 'end'
